#include "header.h"

extern char *optarg;
extern int optind;

/*
 * Enable debugging. Default off.
 */
int debug = 1;

/*
 * Only show info
 */
int info;

/*
 * Size of queue. Default no queueing, only one request at a time
 */
int queued = 1;

/*
 * File descriptor list.
 * Might have to open multiple fd's
 * per device since queue depth per fd
 * is limited to 16
 */
int fd[64];

/*
 * The current queue for a file descriptor
 */
int fdpth[64];

/*
 * Size of the I/O to be performed
 */
int iosize = 1;

/*
 * Do we need to insmod and rmmod?
 * Default is no. Use -k to force.
 */
int module;

/*
 * Number of file descriptors to open per device
 * to fulfil queue depth requirement
 */
int numFd;

/*
 * Total number of writes performed
 */
unsigned long totwrite = 0;

/*
 * Total queue depth
 */
int totdpth;

/*
 * Average queue depth
 */
int avgdpth;

/*
 * Blocksize and capacity of the drive
 */
int blocksize;
int capacity;

/*
 * Increment to the next lba
 */
int increment;

/*
 * Should direct i/o be performed?
 * Default on. Disable with -i
 */
int directio = 1;

/*
 * Page size of system
 */
int psz;

/*
 * Main dxferp and sbp area
 */
char *dxfer;
char *sb;

/*
 * Mode of operation
 *
 * modeSR = 0 => Sequential (Default)
 * modeSR = 1 => Random
 *
 * modeRW = 0 => Read (Default)
 * modeRW = 1 => Write 
 */
int modeSR = 0;
int modeRW = 0;
int mixed = 0;
int mixedP = 0;

/*
 * Aligned malloc ptr
 */
void *alloc_bp1;
void *alloc_bp2;

/*
 * Which device to test
 */
char *device = NULL;

/*
 * The command block for the write() and read()
 */
char cmdpr[10] = { READ_10, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
char cmdpw[10] = { WRITE_10, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

/*
 * Header for read() in a sigPoll
 */
sg_io_hdr_t bufr;

/*
 * Header for a write()
 */
sg_io_hdr_t bufw;

int main( int argc, char *argv[] )
{
  /*
   * Amount of time in seconds to run
   * Default = 30 seconds
   */
  int timerun = 30;
  time_t starttime;

  /*
   * getopt() variable
   */
  char opt;

  /*
   * Counter
   */
  int i;

  /*
   * Total data transferred
   */
  double totaltrans;

  /*
   * Command for enabling directio
   */
  char command[255];

  /*
   * Get the page size of the system
   */
  psz = getpagesize();

  /*
   * Commandline options
   */
  while( ( opt = getopt( argc, argv, "d:kigq:t:s:hrwnm:" ) ) != -1 )
  {
    switch(opt)
    {
      case 'd':
        /* 
	 * Which device to open
	 * Eg: /dev/sg3
	 * If not specified, program prints a list of
	 * available devices and asks user for input.
	 * (interactive)
	 */
        device = (char *) malloc( strlen( optarg ) * sizeof( char ) );
        strncpy( device, optarg, strlen( optarg) );
	break;
      case 'k':
        /*
	 * Force insert of module sg.o in the kernel
	 */
        module = 1;
	break;
      case 'i':
        directio = 0;
        break;
      case 'g':
        debug = 0;
        break;
      case 'q':
        /*
         * Queueing requested. Make sure parameter is number.
         */
        if( isnumber( optarg ) ) queued = atoi( optarg );
        else 
        {
          printf( "Invalid argument %s for -q\n", optarg );
          printHelp();
        }
        break;
      case 't':
        /*
	 * Amount of time in seconds to run
	 */
	if( isnumber( optarg ) ) timerun = atoi( optarg );
	else
	{
	  printf( "Invalid argument %s for -t\n", optarg );
	  printHelp();
	}
	break;
      case 's':
        /*
         * I/O size
         */
	if( isnumber( optarg ) ) iosize = atoi( optarg ) * 1024;
	else
	{
	  printf( "Invalid argument %s for -s\n", optarg );
	  printHelp();
        }
        break;
      case 'r':
        /* 
         * Change mode to random
         * Seq = 0; Ran = 1
         */
        modeSR = 1;
        break;
      case 'w':
        /*
         * Change mode to write
         * Read = 0; Write = 1
         */
        modeRW = 1;
        break;
      case 'n':
        /*
         * Show device info and quit
         */
        info = 1;
        break;
      case 'm':
        /*
         * Mixed mode of reads and writes based on the % provided
         *
         * If argument = 33, then 33% reads and 67% writes.
         */
        mixed = 1;
        if( isnumber( optarg ) )
        {
          if( atoi( optarg ) < 100 )
            mixedP = atoi( optarg );
          else
          {
            printf( "Argument of -m should be between 0 and 100\n" );
            printHelp();
          }
        }
        else
        {
          printf( "Invalid argument %s for -m\n", optarg );
          printHelp();
        }
        break;
      case ':':
        printf( "Parameter missing\n" );
      case 'h':
      case '?':
        printf( "-------------\n"
                "lsperf v%s\n"
                "-------------\n"
                , VERSION );
	printHelp();
	break;
    }
  }

  if( debug ) printf( "-------------\n"
                      "lsperf v%s\n"
                      "-------------\n"
                      , VERSION );

  /*
   * Insert sg.o into the kernel
   */
  if( module ) insertModule();
  
  /* 
   * Show the devices in /proc/scsi/sg/device_str
   * and allow user to select one if device is
   * not specified on cmdline with -d
   */
  if( device == NULL ) device = deviceSel( device );
  if( debug ) printf( "Device selected: %s\n", device );

  /*
   * Set the i/o mode requested in /proc
   */
  sprintf( command, "echo \"%d\" > /proc/scsi/sg/allow_dio", directio );

  if( WEXITSTATUS( system(command) ) )
  {
    printf( "Failed to set i/o mode requested.\n"
            "Is sg.o present in the kernel? See -k.\n\n" );
    printHelp();
  }
  if( debug ) printf( "Enabled %s I/O mode.\n\n", (directio)?"Direct":"Indirect" );

  /*
   * Get device information like queue depth, capacity, etc
   */
  devinfo( device );

  /*
   * Final queue size
   */
  if( debug ) 
    printf( "Requested queue depth = %d\n"
            "I/O Size = %dk\n", queued, iosize / 1024 );

  /*
   * Calculate the increment to the next lba
   */
  increment = iosize/blocksize;

  /*
   * Display what mode we are in
   */
  if( debug )
  {
    if( !mixed )
      printf( "Requested mode: %s %s\n", 
	    (!modeSR)?"Sequential":"Random",
	    (!modeRW)?"Read":"Write" );
    else
      printf( "Requested mode: %s Read/Write (%d%%:%d%%)\n", 
	    (!modeSR)?"Sequential":"Random",
	    mixedP, 100-mixedP );
  }

  /*
   * Number of file descriptors to open
   */
  numFd = queued / SG_MAX_QUEUE;
  if( queued % SG_MAX_QUEUE ) numFd++;

  /*
   * If only info requested, quit.
   */
  if( info ) 
  {
    removeModule( module );
    exit(EXIT_SUCCESS);
  }

  if( debug ) printf( "Opening %d file descriptors\n\n", numFd );

  /*
   * Open the file descriptors with read/write permission
   * and in non blocking mode to allow for command queueing
   */
  for( i = 0; i < numFd; i++ )
  {
    fd[i] = open( device, O_RDWR | O_NONBLOCK );
    if( fd[i] == -1 )
    {
      perror( "open()" );
      removeModule( module );
      exit(EXIT_FAILURE);
    }
    fdpth[i] = 0;
  }

  /*
   * Set up the sigPoll and Ctrl-C related signal action
   */
  setSignal();
  
  /*
   * malloc() space for dxfer and sb area depending on queue
   * size and iosize 
   */
  if( (alloc_bp1 = malloc( ( sizeof( char ) * iosize * queued ) + psz ) ) == NULL )
  {
    perror( "malloc()" );
    removeModule( module );
    exit(EXIT_FAILURE);
  }
  else
  {
    dxfer = (char *) (((unsigned long)alloc_bp1 + psz - 1) & (~(psz - 1)));
    if( dxfer == NULL )
    {
      printf( "Error with aligned malloc()\n" );
      removeModule( module );
      exit(EXIT_FAILURE);
    }
  }
  if( (alloc_bp2 = malloc( ( sizeof( char ) * 32 * queued ) + psz ) ) == NULL )
  {
    perror( "malloc()" );
    removeModule( module );
    exit(EXIT_FAILURE);
  }
  else
  {
    sb = (char *) (((unsigned long)alloc_bp2 + psz - 1) & (~(psz - 1)));
    if( sb == NULL )
    {
      printf( "Error with aligned malloc()\n" );
      removeModule( module );
      exit(EXIT_FAILURE);
    }
  }

  /*
   * Set permanent values in cmd blocks
   */
  cmdpr[8] = iosize/blocksize;
  cmdpw[8] = iosize/blocksize;

  /*
   * Initialize the sg_io_hdr_t blocks
   */
  memset( &bufw, 0, sizeof( sg_io_hdr_t ) );
  memset( &bufr, 0, sizeof( sg_io_hdr_t ) );
  sgBufInit( &bufw );
  sgBufInit( &bufr );

  /*
   * Set direct I/O flag if requested
   */
  if( directio )
    bufw.flags = SG_FLAG_DIRECT_IO;
  else 
    bufw.flags = 0;

  /*
   * Start the stress, one write per fd at a time and maintain
   * queue full
   */
  starttime = time(0);

  /*
   * Write commands to the file descriptors opened above.
   * Reads will be performed when the fd is ready for a read
   * and a signal SIGRTMIN is generated.
   */
  if( mixed )
  {
    /* 
     * If mixed mode selected
     */
    int p;

    while( ( time(0) - starttime ) < timerun )
    {
      for( i = 0; i < numFd; i++ )
      {
	/*
	 * Randomize reads vs writes if mixed mode selected
	 */
	p = rand() % 100;

	if( p <= mixedP ) workr( i );
	else workw( i );

	/*
	 * Check to see if amount of time to run has been exceeded.
	 */
	if( ( time(0) - starttime ) > timerun ) break;
      }
    }
  }
  else
  {
    if( !modeRW )
    {
      /*
       * Read selected
       */
      while( ( time(0) - starttime ) < timerun )
      {
	for( i = 0; i < numFd; i++ )
	{
	  workr( i );

	  /*
	   * Check to see if amount of time to run has been exceeded.
	   */
	  if( ( time(0) - starttime ) > timerun ) break;
	}
      }
    }
    else
    {
      /*
       * Write selected
       */
      while( ( time(0) - starttime ) < timerun )
      {
	for( i = 0; i < numFd; i++ )
	{
	  workw( i );

	  /*
	   * Check to see if amount of time to run has been exceeded.
	   */
	  if( ( time(0) - starttime ) > timerun ) break;
	}
      }
    }
  }

  /*
   * Print average queue depth
   */
  if( debug ) printf( "Average queue depth = %ld\n\n", avgdpth / totwrite );

  /*
   * How much data transferred. Calculations.
   */
  totaltrans = (double)totwrite * (double)iosize;

  /*
   * Print results
   */
  if( debug )
  {
    printf( "Total tranfer = %f bytes\n"
            "Time taken = %d sec\n"
	    "Data transfer rate = %.2f bytes/sec\n", 
            totaltrans,
	    timerun,
	    totaltrans/timerun );
  }
  else
  {
    /* 
     * Print csv format test condition variables as follows:
     *
     * I/O Method, I/O Mode, Requested Queue Depth, I/O Size(k), 
     * Read/Write Distribution (%), Total Transfer(bytes), 
     * Time Taken(sec), Transfer Rate(bytes/sec)
     *
     * Eg: "Direct","Sequential Read",64,64,0,655360,10,65536
     * Eg: "Indirect","Random Read/Write",32,16,33,32768,1,32768
     */
    printf( "\"%s\",\"%s %s\",%d,%d,%d,%.0f,%d,%.2f\n", 
      (directio)?"Direct":"Indirect",
      (modeSR)?"Random":"Sequential",
      (!mixed)?((modeRW)?"Write":"Read"):"Read/Write",
      queued,
      iosize/1024,
      mixedP,
      totaltrans,
      timerun,
      totaltrans/timerun );
  }

  /*
   * Remove sg.o from kernel
   */
  removeModule( module );

  return EXIT_SUCCESS;
}

/*
 * Insert sg.o into the kernel for the program to work 
 * in the first place.
 */
void insertModule()
{
  char command[255] = "insmod sg";

  if( WEXITSTATUS( system(command) ) ) 
  {
    printf( "Error loading module sg.o\n" );
    exit(EXIT_FAILURE);
  }

  if( debug ) printf( "Insmod successful\n" );
}

/*
 * Remove the module from the kernel. Make sure open file
 * descriptors are closed before doing so.
 */
void removeModule( int flg )
{
  char command[255] = "rmmod sg";
  int i;

  /*
   * Wait for a short while for things to settle down
   */
  sleep(4);

  /*
   * Stop counting reads. Wait till no outstanding read requests remain
   */
  outstanding();

  /*
   * Close the open file descriptors
   */
  for( i = 0; i < numFd; i++ ) close( fd[i] );

  /*
   * If we need to remove the module
   */
  if( flg )
  {
    if( WEXITSTATUS( system(command) ) )
    {
      printf( "Error removing module sg.o\n" );
      exit(EXIT_FAILURE);
    }

    if( debug ) printf( "Rmmod successful\n" );
  }

  free( device );
  free( alloc_bp1 );
  free( alloc_bp2 );
}

/*
 * List of command-line arguments.
 */
void printHelp()
{
  printf( "Program expects sg.o version 3.1.19 or greater\n"
          "to be present in kernel [insmod/kerneld]. See -k.\n\n"
          "Flags:\n"
          "  -d <devicename> : Select (d)evice to test\n"
          "                    Eg: -d /dev/sg3\n"
          "  -g              : Disable debu(g)ging\n"
          "  -h              : This (h)elp menu\n"
          "  -i              : Set I/O mode to (I)ndirect              [Direct I/O]\n"
	  "  -k              : Used to load/unload sg.o into\n"
	  "                    (k)ernel manually.\n"
          "  -m %%of reads    : (M)ixed mode of R/W\n"
          "                    Eg: 33 gives 33%% reads, 67%% writes\n"
          "  -n              : Show device i(n)fo and quit\n"
          "  -q #depth       : Set (q)ueue depth                       [1]\n"
          "                    Maximum device queue length available\n"
          "                    at runtime. Check -f.\n"
          "  -r              : (R)andom testing                        [Sequential]\n"
          "  -s #iosize      : Set I/O (s)ize.                         [1]\n"
          "                    Eg: -s  1 =>  1k I/O size\n"
          "                        -s 64 => 64k I/O size\n"
	  "  -t #seconds     : Amount of (t)ime to run test            [30]\n"
          "  -w              : Perform (w)rite test                    [Read]\n" );

  exit(EXIT_FAILURE);
}

/*
 * Check if a command-line argument is purely numbers and not alphanumeric.
 */
int isnumber( char *arg )
{
  int ctr;

  int len = strlen( arg );
          
  for( ctr = 0; ctr < len; ctr++ )
    if( !isdigit( arg[ctr] ) ) break;

  if( ctr != len ) return 0;

  return 1;
}

/*
 * Catch SIGINT and SIGTERM to quit gracefully.
 * Also, set up the SIGRTMIN signal to call sigPoll()
 * when a file descriptor is ready to do a read().
 * Hence, command queueing achieved without poll().
 */
void setSignal()
{
  int i;
  int flags;
  struct sigaction sa;

  sa.sa_flags = SA_SIGINFO;
  sa.sa_sigaction = sigPoll;
  sigaction( SIGRTMIN, &sa, 0 );

  for( i = 0; i < numFd; i++ )
  {
    fcntl( fd[i], F_SETOWN, getpid() );
    flags = fcntl( fd[i], F_GETFL );
    fcntl( fd[i], F_SETFL, flags | O_ASYNC );
    fcntl( fd[i], F_SETSIG, SIGRTMIN );
  }

  signal( SIGINT, catchC );
  signal( SIGTERM, catchC );
}

/*
 * Catch premature termination and quit gracefully.
 */
void catchC()
{
  if( debug ) printf( "Control-C caught\n" );
  
  /*
   * Close file descriptors and remove module
   */
  removeModule( module );

  exit(EXIT_FAILURE);
}
  
void outstanding()
{
  int i;
  sg_io_hdr_t buf;

  while( totdpth != 0 )
    for( i = 0; i < numFd; i++ )
      while( fdpth[i] != 0 )
	if( read( i, &buf, sizeof( sg_io_hdr_t ) ) > 0 )
        {
          fdpth[i]--;
	  totdpth--;
        }
}
