#include "header.h"

/*
 * External variables
 */
extern int fd[16];
extern int fdpth[16];
extern int iosize;
extern int numFd;
extern int module;
extern int queued;
extern int totdpth;
extern int avgdpth;
extern int blocksize;
extern int capacity;
extern int directio;
extern int totwrite;
extern char *dxfer;
extern char *sb;
extern int modeSR;
extern char cmdpr[10];
extern char cmdpw[10];
extern sg_io_hdr_t bufr;
extern sg_io_hdr_t bufw;
extern int increment;

/*
 * Which Logical Block Address to read from or write from.
 */
static unsigned int lba;
static int ID = 0;

/*
 * sgBufInit function used to initialize sg_io_hdr_t
 * structs.
 * If i = 0; initialize as write()
 * If i = 1; initialize as read()
 */
void sgBufInit( sg_io_hdr_t *buf )
{
  buf->interface_id = 'S';

  /*
   * Command length (Use only 10 byte READ and WRITE)
   */
  buf->cmd_len = 10;

  /*
   * Sense code return length
   */
  buf->mx_sb_len = 32;

  /*
   * Scatter gather not usable in DIRECT_IO
   */
  buf->iovec_count = 0;

  /*
   * Amount of data to move
   */
  buf->dxfer_len = iosize; 

  /*
   * Timeout, set to 3 seconds
   */
  buf->timeout = 3000;
}

/*
 * sgBufSet function to fix lba address and data direction
 */
void sgBufSet( sg_io_hdr_t *buf )
{
  /*
   * Set the lba address
   */
  buf->cmdp[2] = (unsigned char) ( ( lba & 0xff000000 ) >> 24 );
  buf->cmdp[3] = (unsigned char) ( ( lba & 0xff0000 ) >> 16 );
  buf->cmdp[4] = (unsigned char) ( ( lba & 0xff00 ) >> 8 );
  buf->cmdp[5] = (unsigned char) ( lba & 0xff );

  /*
   * New lba based on current mode.
   */
  if( !modeSR )
    lba = (lba + ( increment )) % capacity;
  else
    lba = rand() % capacity;
}

void sigPoll( int sig, siginfo_t *si, void *data )
{
  /*
   * Which file descriptor is ready for a read
   */
  int off = offFd( si->si_fd );

  /*
   * Read the result of the command
   */
  read( si->si_fd, &bufr, sizeof( sg_io_hdr_t ) );

  /*
   * Check status of command
   */
  if( bufr.masked_status != GOOD )
  {
    printf( "Write failed, Status: 0x%2x\n", bufr.masked_status );
    removeModule( module );
    exit(EXIT_FAILURE);
  }

  /*
   * Check if direct I/O was performed
   */
  if( !( bufr.info & SG_INFO_DIRECT_IO_MASK ) )
    if( directio )
    {
      printf( "Direct I/O failed.\n" );
      removeModule( module );
      exit(EXIT_FAILURE);
    }

  /*
   * Check if the file descriptor in siginfo_t is valid
   */
  if( off == -1 )
  {
    printf( "Fatal error with sigPoll()\n" );
    removeModule( module );
    exit(EXIT_FAILURE);
  }

  /*
   * Decrement current fd queue usage
   */
  fdpth[off]--;

  /*
   * Decrement total queue usage
   */
  totdpth--;

  /*
   * Increment total packets written
   */
  totwrite++;
}

/*
 * Function to return the offset for the filedescriptor
 * specified from fd[]
 */
int offFd( int d )
{
  int i;

  for( i = 0; i < numFd; i++ )
    if( d == fd[i] ) return i; 

  return -1;
}

/*
 * Do a write() onto the SCSI bus. Command is a READ_10
 */
void workr( int off )
{
  /*
   * Check if this fd has reached its max queue depth
   * Check if total queue depth has been exceeded
   */
  if( fdpth[off] == SG_MAX_QUEUE || totdpth == queued )
    return;

  /*
   * Assign the pointer values and the ID of the packet
   */
  bufw.cmdp = cmdpr;
  bufw.dxferp = dxfer + ( ID * iosize );
  bufw.sbp = sb + ( ID * 32 );
  bufw.dxfer_direction = SG_DXFER_FROM_DEV;

  /*
   * Initialize the sg_io_hdr_t header.
   */
  sgBufSet( &bufw );

  /*
   * Increment ID
   */
  ID = (ID+1) % queued;

  /*
   * write() to the file descriptor
   */
  if( write( fd[off], &bufw, sizeof( sg_io_hdr_t ) ) < 0 )
  {
    perror( "write()" );
    removeModule( module );
    exit(EXIT_FAILURE);
  }

  /*
   * Each file descriptor has a queue depth of 16. Hence
   * keep track of how many writes have been done per fd
   * and decrement when a read() has been performed from
   * this fd. (Done in sigPoll())
   */
  fdpth[off]++;

  /*
   * Total queue depth. This is to keep track of the overall
   * queue usage. Relevent only if multiple fd's opened per
   * device.
   */
  totdpth++;

  /*
   * Keep track of average queue depth
   */
  avgdpth += totdpth;
}
  
/*
 * Do a write() onto the SCSI bus. Command is a WRITE_10
 */
void workw( int off )
{
  /*
   * Check if this fd has reached its max queue depth
   * Check if total queue depth has been exceeded
   */
  if( fdpth[off] == SG_MAX_QUEUE || totdpth == queued )
    return;

  /*
   * Assign the pointer values and the ID of the packet
   */
  bufw.cmdp = cmdpw;
  bufw.dxferp = dxfer + ( ID * iosize );
  bufw.sbp = sb + ( ID * 32 );
  bufw.dxfer_direction = SG_DXFER_TO_DEV;

  /*
   * Initialize the sg_io_hdr_t header.
   */
  sgBufSet( &bufw );

  /*
   * Increment ID
   */
  ID = (ID+1) % queued;

  /*
   * write() to the file descriptor
   */
  if( write( fd[off], &bufw, sizeof( sg_io_hdr_t ) ) < 0 )
  {
    perror( "write()" );
    removeModule( module );
    exit(EXIT_FAILURE);
  }

  /*
   * Each file descriptor has a queue depth of 16. Hence
   * keep track of how many writes have been done per fd
   * and decrement when a read() has been performed from
   * this fd. (Done in sigPoll())
   */
  fdpth[off]++;

  /*
   * Total queue depth. This is to keep track of the overall
   * queue usage. Relevent only if multiple fd's opened per
   * device.
   */
  totdpth++;

  /*
   * Keep track of average queue depth achieved
   */
  avgdpth += totdpth;
}
